/* information on how to use the picture routines can be found in pic.h !! */
#include "pic.h"
#include "sys/time.h"
#include "omp.h"
/* to deal with complex numbers */
typedef struct {
        float real;
        float imag;
} complex;

/* a piecewise linear bump with maximal height 255 */
Pixel bump(float s)
{
        float t;

        /* s can be in [-1,1] */
        if (s<0) t = (s+1.0); else t = s;

        t = 6*t;


        if (t <= 1) return (Pixel) (t*255);
        if (t <= 3) return (Pixel) 255;
        if (t <= 4) return (Pixel) 255*(4.0-t);
        return (Pixel) 0;
}

void palette(float colorval, Pixel *rgb)
{

        float cv, pt;

        if (colorval > 0.999)
        {
		/* many mandelbrot iterations end up here (white) */
                rgb[2] = (Pixel) 64;
                rgb[1] = (Pixel) 64;
                rgb[0] = (Pixel) 64;
        }
        else
        {
		/* distribute values from 0 to pt over the spectrum */
                pt = 0.06;
	        if (colorval<pt) cv = colorval/pt; else cv = 1.0;
                rgb[2] = bump(cv);
                rgb[1] = bump(cv-1.0/3.0);
                rgb[0] = bump(cv-2.0/3.0);
        }
}


/* do the mandelbrot iteration */
void cal_pixel(complex c, Pixel *rgb)
{
  int count, max;
  complex z;
  float temp, lengthsq, lma;

  /* maximal number of iterations */
  max = 1024; lma = 700;
  /* initial value for iteration */
  z.real = 0.0;
  z.imag = 0.0;
  /* iteration counter */
  count = 0;
  do
  {
    /* iteration z^2+c->z */
    temp = z.real * z.real - z.imag * z.imag +c.real;
    z.imag = 2 * z.real * z.imag + c.imag;
    z.real = temp;
    /* square of modulus */
    lengthsq = z.real * z.real + z.imag * z.imag;
    count++;
  }  while ((lengthsq < lma) && (count < max));

  /* calculate color depending on count */
  palette(count/((float)max), rgb);
}

/* calculate a line */
void process_line(float xmin, float xmax, float d, int nx, int y, Pic *pcol)
{
  	int x, chan;
	complex c;
	Pixel rgb[3];

	/* y coordinate fixed in line */
        c.imag = xmax - (y+0.5)*d;

	for (x=0; x<nx; x++)
        {
          /* coordinate of current point */
          c.real = xmin + (x+0.5)*d;
	  cal_pixel(c, rgb);
          for (chan=0; chan<pcol->bpp; chan ++)
             PIC_PIXEL(pcol, x, 0, chan) = rgb[chan];
        }
}

/* put line into final picture */
incorporate(int nx, int y, Pic *pcol, Pic *pout){
  int x, chan;
  for (x=0; x<nx; x++){
    for (chan=0; chan<pout->bpp; chan ++)
      PIC_PIXEL(pout, x, y, chan) = PIC_PIXEL(pcol, x, 0, chan);
  }
}


int main(int argc, char **argv)
{
  	struct timeval start, end;
	/* number of pixels */
	int nx;
	/* colored picture */
	Pic *pout, *pcol; 
	/* loop variable */
	int y;
	/* x-interval of quadratic picture */
	float xmin=-2.0, xmax=2.0;
	/* thickness of pixel */
	float d;
	
	
	/* get command line options */
	if (argc<3){
	  printf("need arguments: number_of_pixels outpufile.pnm\n");
	  exit(1);
	}

	/* reading number of pixels in x direction */
        if (sscanf(argv[1], "%d", &nx) != 1){
          printf("error, got %s instead of an integer", argv[1]);
          exit(1);
        }
        if (nx < 1){
          printf("need at least one pixel, aborting\n");
          exit(1);
        }

        /* thickness of pixel */
        d = (xmax - xmin)/nx;

	/* allocate space for final picture */
	pout = pic_alloc(nx, nx, 3);
	
	int nthreads = 8;
	
	gettimeofday(&start, 0);
	/*start of parallel block*/
#pragma omp parallel private(y,pcol) num_threads(nthreads)
	{
	  /* allocate space for one line */
	  pcol = pic_alloc(nx, 1, 3);
#pragma omp for schedule(dynamic,1)
	  for(y = 0; y < nx; y++){
	    process_line(xmin, xmax, d, nx, y, pcol);
	    incorporate(nx, y, pcol, pout);
	  }
	  /* removing picture */
	  pic_free(pcol);
	}
	/*end of parallel block*/
	gettimeofday(&end, 0);
	double timeElapsed = (double)(end.tv_sec - start.tv_sec) + ((double)end.tv_usec/1000000 - (double)start.tv_usec/1000000);
		printf("Anzahl Threads: %d Zeit: %f sec\n" ,nthreads,timeElapsed);
		

	/* write pnm file */
        if (! pnm_write(argv[2], pout))
        {
          printf("error writing %s\n",argv[2]);
          exit(1);
	  }
	/* removing picture */
	pic_free(pout);
	
	exit(0);
}
